from logging.config import fileConfig

import sqlalchemy as sa
from alembic import context

from eztea.sql._connection import SqlalchemyConnection


class AlembicEnvPy:
    def __init__(
        self,
        metadata: sa.MetaData,
        connection: SqlalchemyConnection,
        **kwargs,
    ) -> None:
        self._metadata = metadata
        self._connection = connection
        self._kwargs = kwargs

    def run_migrations_offline(self):
        """Run migrations in 'offline' mode.

        This configures the context with just a URL
        and not an Engine, though an Engine is acceptable
        here as well.  By skipping the Engine creation
        we don't even need a DBAPI to be available.

        Calls to context.execute() here emit the given string to the
        script output.
        """
        context.configure(
            url=self._connection.url,
            target_metadata=self._metadata,
            literal_binds=True,
            dialect_opts={"paramstyle": "named"},
            **self._kwargs,
        )
        with context.begin_transaction():
            context.run_migrations()

    def run_migrations_online(self):
        """Run migrations in 'online' mode.

        In this scenario we need to create an Engine
        and associate a connection with the context.
        """
        with self._connection.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=self._metadata,
                **self._kwargs,
            )
            with context.begin_transaction():
                context.run_migrations()

    def run_if_alembic(self):
        # this is the Alembic Config object, which provides
        # access to the values within the .ini file in use.
        config = getattr(context, "config", None)
        if config is None:
            return
        # Interpret the config file for Python logging.
        # This line sets up loggers basically.
        fileConfig(config.config_file_name)
        if context.is_offline_mode():
            self.run_migrations_offline()
        else:
            self.run_migrations_online()



from typing import Any, Optional, Union

import falcon
from validr import Invalid

__all__ = (
    "BaseWebError",
    "RequestParamsInvalid",
    "ResponderReturnsInvalid",
)


class BaseWebError(Exception):
    """BaseWebError"""

    message: str
    error: str
    detail: Optional[Any]
    status: falcon.HTTPStatus
    headers: Optional[Union[list, dict]]

    def __init__(
        self,
        message: str,
        error: str = None,
        detail: Any = None,
        status: falcon.HTTPStatus = falcon.HTTP_400,
        headers: Union[list, dict] = None,
    ) -> None:
        self.message = message
        self.error = error or type(self).__name__
        self.detail = detail
        self.status = status
        self.headers = headers

    def __repr__(self):
        type_name = type(self).__name__
        error = ""
        if self.error != type_name:
            error = f"{self.error}: "
        return f"<{type_name} {error}{self.message}>"

    def __str__(self) -> str:
        return f"{self.error}: {self.message}"


class RequestParamsInvalid(BaseWebError):
    """RequestParamsInvalid"""

    def __init__(self, error: Invalid) -> None:
        value = error.value
        if value is not None:
            if not isinstance(value, (bool, int, float, str)):
                value = repr(value)
        detail = {
            "field": error.position,
            "value": value,
            "issue": error.message,
        }
        super().__init__(str(error), detail=detail)


class ResponderReturnsInvalid(Exception):
    """ResponderReturnsInvalid"""

    def __init__(self, error: Invalid) -> None:
        super().__init__(str(error))



import datetime
import decimal
import json
import uuid
from json import JSONDecodeError, JSONDecoder, load, loads

from .clock import format_iso8601 as _format_iso8601
from .clock import is_timezone_aware as _is_timezone_aware

__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
    "JSONDecoder",
    "JSONDecodeError",
    "JSONEncoder",
]


class JSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time, decimal types, and
    UUIDs.

    source: https://github.com/django/django/blob/3.2.11/django/core/serializers/json.py
    """

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime.datetime):
            return _format_iso8601(o)
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if _is_timezone_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, (decimal.Decimal, uuid.UUID)):
            return str(o)
        else:
            return super().default(o)


def dumps(obj, ensure_ascii=False, indent=None, **kwargs):
    """
    Serialize ``obj`` to a JSON formatted ``str``.
    """
    return json.dumps(
        obj,
        indent=indent,
        ensure_ascii=ensure_ascii,
        cls=JSONEncoder,
        **kwargs
    )


def dump(obj, fp, ensure_ascii=False, indent=None, **kwargs):
    """
    Serialize ``obj`` as a JSON formatted stream to ``fp`` (a
    ``.write()``-supporting file-like object).
    """
    json.dump(
        obj,
        fp,
        indent=indent,
        ensure_ascii=ensure_ascii,
        cls=JSONEncoder,
        **kwargs
    )



import cgi

import falcon
from validr import T, modelclass

__all__ = (
    "MIME_TYPE_JSON",
    "MIME_TYPE_MULTIPART",
    "MIME_TYPE_URLENCODED",
    "MIME_TYPE_TEXT",
    "parse_content_type",
    "Request",
    "FormFile",
)

MIME_TYPE_JSON = "application/json"
MIME_TYPE_MULTIPART = "multipart/form-data"
MIME_TYPE_URLENCODED = "application/x-www-form-urlencoded"
MIME_TYPE_TEXT = "text/plain"


def parse_content_type(content_type: str):
    if not content_type:
        return ("", {})
    mimetype, params = cgi.parse_header(content_type)
    return mimetype.lower(), params


class Request(falcon.Request):
    @property  # TODO: remove py37 and use cached_property
    def __content_type_parsed(self):
        return parse_content_type(self.content_type)

    @property
    def mimetype(self):
        """
        Like content_type, but without parameters (eg, without charset,
        type etc.) and always lowercase. For example if the content type is
        text/HTML; charset=utf-8 the mimetype would be 'text/html'.
        """
        return self.__content_type_parsed[0]

    @property
    def mimetype_params(self):
        """
        The mimetype parameters as dict. For example if the content type is
        text/html; charset=utf-8 the params would be {'charset': 'utf-8'}.
        """
        return self.__content_type_parsed[1]


@modelclass
class FormFile:
    # content type
    content_type: str = T.str
    # secure filename
    filename: str = T.str
    # file data
    data: bytes = T.bytes



from fnmatch import fnmatch
from typing import Any, List

from sqlalchemy.schema import SchemaItem


class IgnoreTables:
    """
    Ignore tables from migration.

        AlembicEnvPy(
            include_object=IgnoreTables(["ignore_*", "*_ignored"]),
        )

    """

    def __init__(self, ignore_tables: List[str]) -> None:
        self._ignore_tables = ignore_tables or []

    def _include_object(
        self,
        object: SchemaItem,
        name: str,
        type_: str,
        reflected: bool,
        compare_to: Any,
    ):
        """
        Should you include this table or not?
        https://stackoverflow.com/questions/65184035/alembic-ignore-specific-tables
        """
        if object.info.get("skip_autogenerate", False):
            return False
        if type_ == "table":
            if any(fnmatch(name, p) for p in self._ignore_tables):
                return False
        return True

    __call__ = _include_object



import os
import os.path
from typing import Type, TypeVar

from dotenv import load_dotenv
from validr import fields

__all__ = ("load_env_config",)

T = TypeVar("T")


def load_env_config(
    *,
    model_type: Type[T],
    env_prefix: str,
    default_envfile: str = None,
) -> T:
    """
    Load envfile and convert to config model type.

    Args:
        model_type: validr.modelclass type
        env_prefix: env variable key prefix, eg: 'MY_APP_'
        default_envfile: default envfile path, optional
    """
    envfile_path = os.getenv(f"{env_prefix}CONFIG")
    if not envfile_path:
        if default_envfile and os.path.exists(default_envfile):
            envfile_path = default_envfile
    if envfile_path:
        envfile_path = os.path.abspath(os.path.expanduser(envfile_path))
        print(f"* Load envfile at {envfile_path}")
        load_dotenv(envfile_path)
    configs = {}
    for name in fields(model_type):
        key = (env_prefix + name).upper()
        value = os.environ.get(key, None)
        if value is not None:
            configs[name] = value
    return model_type(configs)



from collections import ChainMap
from typing import Any, Callable, Dict, List, Optional, Union

import falcon
from falcon.media.multipart import MultipartForm
from validr import Builder, Compiler, Invalid, Schema

from eztea import json

from ._signature import get_params, get_returns
from .error import RequestParamsInvalid, ResponderReturnsInvalid
from .request import (
    MIME_TYPE_JSON,
    MIME_TYPE_MULTIPART,
    MIME_TYPE_URLENCODED,
    Request,
    parse_content_type,
)


class RouterHandler:
    def __init__(
        self,
        fn: Callable,
        path: str,
        methods: List[str],
        params: Optional[Union[Schema, Builder]],
        returns: Optional[Union[Schema, Builder]],
        description: Optional[str],
        schema_compiler: Compiler,
    ) -> None:
        self.fn = fn
        self.path = path
        self.methods = methods
        self.params = params
        self.returns = returns
        self.description = description
        self._validate_params = None
        if params is not None:
            self._validate_params = schema_compiler.compile(params)
        self._validate_returns = None
        if returns is not None:
            self._validate_returns = schema_compiler.compile(returns)

    def _extract_params(self, request: Request, kwargs: dict) -> dict:
        data_s = [kwargs]
        media = request.get_media(default_when_empty=None)
        if media is not None:
            if request.mimetype == MIME_TYPE_MULTIPART:
                data_s.append(self._extract_multipart(media))
            else:
                data_s.append(media)
        if request.params is not None:
            data_s.append(request.params)
        request_data = ChainMap(*data_s)
        try:
            params = self._validate_params(request_data)
        except Invalid as ex:
            raise RequestParamsInvalid(ex) from ex
        return params

    def _extract_multipart(self, form: MultipartForm):
        params = {}
        for part in form:
            is_file = bool(part.filename)
            if is_file:
                params[part.name] = dict(
                    filename=part.secure_filename,
                    content_type=part.content_type,
                    data=part.get_data(),
                )
            else:
                mimetype, _ = parse_content_type(part.content_type)
                if mimetype in (MIME_TYPE_URLENCODED, MIME_TYPE_JSON):
                    params[part.name] = part.get_media()
                else:
                    params[part.name] = part.get_text()
        return params

    def on_request(
        self,
        request: Request,
        response: falcon.Response,
        **kwargs,
    ):
        if self._validate_params is not None:
            params = self._extract_params(request, kwargs)
        else:
            # fallback to falcon usage with path match params
            params = kwargs
        ctx = ResponderContext(
            request=request, response=response, params=params
        )
        returns = self.fn(ctx, **params)
        if self._validate_returns is not None:
            try:
                returns = self._validate_returns(returns)
            except Invalid as ex:
                raise ResponderReturnsInvalid(str(ex)) from ex
        if returns is not None:
            response.text = json.dumps(returns)
            response.status = falcon.HTTP_200
            response.content_type = falcon.MEDIA_JSON


class ResponderContext:
    def __init__(
        self,
        request: Request,
        response: falcon.Response,
        params: Dict[str, Any] = None,
    ) -> None:
        self.request = request
        self.response = response
        self.params = params


class RouterResource:
    def __init__(self, path: str, action_s: Dict[str, RouterHandler]) -> None:
        self.path = path
        for method, handler in action_s.items():
            setattr(self, f"on_{method.lower()}", handler.on_request)


_ALL_HTTP_METHODS = tuple(falcon.HTTP_METHODS)


def _normalize_methods(methods: Optional[List[str]]):
    methods = [x.upper() for x in methods or _ALL_HTTP_METHODS]
    unknown_methods = set(methods) - set(_ALL_HTTP_METHODS)
    if unknown_methods:
        msg = ", ".join(unknown_methods)
        raise ValueError(f"unknown http method {msg}")
    return methods


class Router:
    def __init__(self, *, schema_compiler: Compiler = None) -> None:
        self._schema_compiler = schema_compiler or Compiler()
        self._resource_define_s: Dict[str, Dict[str, RouterHandler]] = {}

    def _add_handler(self, handler: RouterHandler):
        action_define_s = self._resource_define_s.setdefault(handler.path, {})
        for method in handler.methods:
            if method in action_define_s:
                msg = f"duplicated route {method} {handler.path}"
                raise ValueError(msg)
            action_define_s[method] = handler

    def to_resource_s(self) -> List[RouterResource]:
        resource_s = []
        for path, action_define_s in self._resource_define_s.items():
            resource_s.append(RouterResource(path, action_define_s))
        return resource_s

    def route(
        self,
        path: str,
        methods: List[str] = None,
        params: Union[Schema, Builder] = None,
    ):
        methods = _normalize_methods(methods)

        def decorator(fn):
            fn_params = params
            if fn_params is None:
                fn_params = get_params(fn)
            fn_returns = get_returns(fn)
            self._add_handler(
                RouterHandler(
                    fn=fn,
                    path=path,
                    methods=methods,
                    params=fn_params,
                    returns=fn_returns,
                    description=fn.__doc__,
                    schema_compiler=self._schema_compiler,
                )
            )
            return fn

        return decorator

    def get(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["GET"], params=params)

    def post(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["POST"], params=params)

    def put(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["PUT"], params=params)

    def delete(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["DELETE"], params=params)

    def patch(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["PATCH"], params=params)

    def head(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["HEAD"], params=params)

    def options(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["OPTIONS"], params=params)

    def trace(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["TRACE"], params=params)

    def connect(self, path: str, params: Union[Schema, Builder] = None):
        return self.route(path, methods=["CONNECT"], params=params)



from typing import List


def join_url_path(__a: str, *paths: List[str]) -> str:
    """Join url paths, insert '/' if needed."""
    r = __a
    for p in paths:
        if r[-1:] == "/" and p[:1] == "/":
            r += p[1:]
        elif r[-1:] != "/" and p[:1] != "/":
            r += "/" + p
        else:
            r += p
    return r


def shorten(text, width, placeholder="..."):
    """
    >>> shorten('123456789', width=8)
    '12345...'
    >>> shorten('123456789', width=9)
    '123456789'
    >>> shorten(None, width=8) is None
    True
    """
    if not text:
        return text
    if len(text) <= width:
        return text
    return text[: max(0, width - len(placeholder))] + placeholder



from collections import namedtuple
from contextvars import ContextVar

TimingValue = namedtuple("TimingValue", ["count", "cost"])


class SqlalchemyConnectionTiming:
    def __init__(self) -> None:
        self._context = ContextVar(
            self.__class__.__name__,
            default=TimingValue(0, 0.0),
        )

    @property
    def value(self) -> TimingValue:
        return self._context.get()

    def increase_count(self, n: int = 1):
        count, cost = self.value
        self._context.set(TimingValue(count + n, cost))

    def increase_cost(self, t: float):
        count, cost = self.value
        self._context.set(TimingValue(count, cost + t))

    def reset(self):
        self._context.set(TimingValue(0, 0.0))



import falcon
from falcon.media.json import JSONHandler
from falcon.media.multipart import MultipartFormHandler, MultipartParseOptions

from eztea import json

from ._helper import join_url_path
from ._router import Router
from .error import BaseWebError
from .request import Request


def _http_error_serializer(req, resp: falcon.Response, ex: falcon.HTTPError):
    resp.text = json.dumps(
        {
            "error": ex.status[:3],
            "message": ex.title,
            "detail": ex.description,
        }
    )
    resp.content_type = falcon.MEDIA_JSON


def _base_web_error_handler(
    req,
    resp: falcon.Response,
    ex: BaseWebError,
    params,
):
    if ex.headers is not None:
        resp.set_headers(ex.headers)
    resp.text = json.dumps(
        {
            "error": ex.error,
            "message": ex.message,
            "detail": ex.detail,
        }
    )
    resp.status = ex.status
    resp.content_type = falcon.MEDIA_JSON


# 阿里云网关限制，请求体大小: 8M
# 腾讯云网关限制，请求体大小: 16M
DEFAULT_MAX_CONTENT_LENGTH = 8 * 1024 * 1024


class MaxContentLengthMiddleware:
    def __init__(
        self,
        max_content_length: int = DEFAULT_MAX_CONTENT_LENGTH,
    ) -> None:
        self._max_content_length = max_content_length
        n_kb = max_content_length // 1024
        self._description = f"Request content length limited to {n_kb}KB"

    def process_request(self, req: falcon.Request, resp):
        if req.content_length is not None:
            if req.content_length > self._max_content_length:
                raise falcon.HTTPPayloadTooLarge(description=self._description)


class Application(falcon.App):
    def __init__(
        self,
        *args,
        middleware=None,
        max_content_length: int = DEFAULT_MAX_CONTENT_LENGTH,
        **kwargs,
    ):
        kwargs.setdefault("request_type", Request)
        super().__init__(*args, **kwargs)

        self.add_middleware(MaxContentLengthMiddleware(max_content_length))
        if middleware is not None:
            self.add_middleware(middleware)

        multipart_options = MultipartParseOptions()
        multipart_options.max_body_part_buffer_size = max_content_length
        multipart_handler = MultipartFormHandler(multipart_options)
        json_handler = JSONHandler(dumps=json.dumps, loads=json.loads)
        self.req_options.media_handlers.update(
            [
                (falcon.MEDIA_MULTIPART, multipart_handler),
                (falcon.MEDIA_JSON, json_handler),
            ],
        )

        self.set_error_serializer(_http_error_serializer)
        self.add_error_handler(BaseWebError, _base_web_error_handler)

    def include_router(self, router: Router, prefix: str = ""):
        for resource in router.to_resource_s():
            self.add_route(join_url_path(prefix, resource.path), resource)



"""
A datetime helper with UTC timezone by default.
"""
from datetime import (  # noqa: F401
    MAXYEAR,
    MINYEAR,
    date,
    datetime,
    time,
    timedelta,
    timezone,
    tzinfo,
)
from typing import Union

import ciso8601

__all__ = (
    "date",
    "datetime",
    "time",
    "timedelta",
    "timezone",
    "tzinfo",
    "MINYEAR",
    "MAXYEAR",
    "UTC",
    "now",
    "is_timezone_aware",
    "format_iso8601",
    "parse_iso8601",
    "from_timestamp",
)


UTC = timezone.utc


def now() -> datetime:
    """Return now in UTC timezone"""
    return datetime.utcnow().replace(tzinfo=UTC)


def is_timezone_aware(value: Union[datetime, time]) -> bool:
    """
    Determine if a given datetime.datetime is timezone aware.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is not None


def format_iso8601(o: datetime) -> str:
    """
    Format datetime to ISO 8601 as well as ECMA-262 format.

    see also: eztea.json.JSONEncoder
    """
    r = o.isoformat()
    if o.microsecond:
        r = r[:23] + r[26:]
    if r.endswith("+00:00"):
        r = r[:-6] + "Z"
    return r


def parse_iso8601(string: str) -> datetime:
    """
    Parse datetime string and return timezone aware datetime object.
    Timezone default to UTC if not provided in datetime string.
    """
    value = ciso8601.parse_datetime(string)
    if value.tzinfo is None:
        value = value.replace(tzinfo=UTC)
    return value


def from_timestamp(t: Union[int, float]) -> datetime:
    """
    Return datetime of the timestamp in UTC timezone.
    """
    return datetime.utcfromtimestamp(t).replace(tzinfo=UTC)



import os
from pathlib import Path
from typing import Union

from ._error import CommandError


def _find_app_dir(base_dir: Path):
    for name in os.listdir(base_dir):
        subdir = base_dir / name
        check_file = subdir / "migration" / "alembic.ini"
        if check_file.is_file():
            return subdir
    raise CommandError("migration directory not exists")


class MigrationDirectory:
    def __init__(self, app_dir: Path) -> None:
        self._app_dir = app_dir.absolute()

    @classmethod
    def from_app_dir(cls, app_dir: Union[str, Path]):
        return cls(Path(app_dir))

    @classmethod
    def from_base_dir(cls, base_dir: Union[str, Path] = "."):
        return cls(_find_app_dir(Path(base_dir)))

    @property
    def app_dir(self):
        return self._app_dir

    @property
    def app_name(self):
        return self._app_dir.name

    @property
    def migration(self):
        return self.app_dir / "migration"

    @property
    def alembic_ini(self):
        return self.migration / "alembic.ini"

    @property
    def versions(self):
        return self.migration / "versions"

    @property
    def ddl(self):
        return self.migration / "ddl"

    @property
    def env_py(self):
        return self.migration / "env.py"

    @property
    def script_py_mako(self):
        return self.migration / "script.py.mako"

    @property
    def template(self):
        return Path(__file__).parent / "template"

    @property
    def template_alembic_ini(self):
        return self.template / "alembic.ini"

    @property
    def template_env_py(self):
        return self.template / "env.py.txt"

    @property
    def template_script_py_mako(self):
        return self.template / "script.py.mako"



import inspect
from validr import T


def get_params(f: callable):
    sig = inspect.signature(f)
    sig_items = list(sig.parameters.values())
    params_schema = {}
    for p in sig_items[1:]:
        if p.default is not inspect.Parameter.empty:
            params_schema[p.name] = p.default
    if params_schema:
        return T.dict(params_schema).__schema__
    return None


def get_returns(f: callable):
    sig = inspect.signature(f)
    if sig.return_annotation is not inspect.Signature.empty:
        schema = sig.return_annotation
        if schema is not None:
            return T(schema).__schema__
    return None



import sqlalchemy as sa

_default_table_kwargs = dict(
    mysql_engine="InnoDB",
    mysql_charset="utf8mb4",
    mysql_collate="utf8mb4_bin",
)


class TableMetaData(sa.MetaData):
    def define_table(self, name: str, *args, **kwargs) -> sa.Table:
        for k, v in _default_table_kwargs.items():
            kwargs.setdefault(k, v)
        return sa.Table(name, self, *args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eztea', 'eztea.sql', 'eztea.sql.migration', 'eztea.web']

package_data = \
{'': ['*'], 'eztea.sql.migration': ['template/*']}

install_requires = \
['ciso8601>=2.2.0',
 'falcon>=3.0.1',
 'python-dotenv>=0.12.0',
 'sqlalchemy-jsonfield>=1.0.0',
 'sqlalchemy-utc>=0.14.0',
 'sqlalchemy-utils>=0.38.2',
 'sqlalchemy>=1.3.19',
 'validr>=1.2.1']

extras_require = \
{'migration': ['click>=7.1.2',
               'alembic>=1.7.5',
               'black>=21.12b0',
               'mako>=1.1.6',
               'python-dateutil>=2.8.2'],
 'mysql': ['pymysql>=1.0.2'],
 'postgresql': ['psycopg2>=2.9.3']}

entry_points = \
{'console_scripts': ['eztea = eztea.__main__:main']}

setup_kwargs = {
    'name': 'eztea',
    'version': '0.1.3',
    'description': 'EZTea Web Framework',
    'long_description': '# EZTea Web Framework\n\n```\npip install eztea\n```\n\n## Usage\n\n```python\nfrom validr import T\n\nfrom eztea.web import Application, ResponderContext, Router\n\nrouter = Router()\n\n\n@router.get("/")\ndef hello(\n    ctx: ResponderContext,\n    name: str = T.str.default("world"),\n) -> T.dict(hello=T.str):\n    return {"hello": name}\n\n\napp = Application()\napp.include_router(router)\n```\n',
    'author': 'guyskk',
    'author_email': 'guyskk@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/guyskk/eztea',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.4,<4.0',
}


setup(**setup_kwargs)



import io
import logging
import os
import os.path
import sys
from contextlib import contextmanager
from typing import List

from alembic import command as _alembic
from alembic.config import Config as _AlembicConfig
from alembic.util.exc import CommandError as _AlembicCommandError

from eztea import clock

from ._directory import MigrationDirectory
from ._error import CommandError

LOG = logging.getLogger(__package__)


@contextmanager
def _alembic_config_output(directory: MigrationDirectory):
    output = io.StringIO()
    cfg = _AlembicConfig(
        directory.alembic_ini,
        output_buffer=output,
        stdout=output,
    )
    try:
        try:
            yield cfg, output
        finally:
            sys.stdout.write(output.getvalue())
    except _AlembicCommandError as ex:
        raise CommandError(str(ex)) from ex


def init(app: str):
    directory = MigrationDirectory.from_app_dir(app)
    directory.app_dir.mkdir(exist_ok=True)
    directory.migration.mkdir(exist_ok=True)
    directory.versions.mkdir(exist_ok=True)
    directory.ddl.mkdir(exist_ok=True)

    alembic_ini = directory.template_alembic_ini.read_text("utf-8")
    alembic_ini = alembic_ini.replace("<<myapp>>", directory.app_name)
    directory.alembic_ini.write_text(alembic_ini, encoding="utf-8")

    env_py = directory.template_env_py.read_text("utf-8")
    env_py = env_py.replace("<<myapp>>", directory.app_name)
    directory.env_py.write_text(env_py, encoding="utf-8")

    script_py = directory.template_script_py_mako.read_text("utf-8")
    directory.script_py_mako.write_text(script_py, encoding="utf-8")


def auto_revision():
    directory = MigrationDirectory.from_base_dir()
    directory.versions.mkdir(exist_ok=True)
    message = clock.format_iso8601(clock.now())
    with _alembic_config_output(directory) as (cfg, _):
        _alembic.revision(cfg, autogenerate=True, message=message)


def _load_versions(directory: MigrationDirectory) -> List[str]:
    version_s = []
    for version_filepath in directory.versions.glob("*.py"):
        filename, _ = os.path.splitext(version_filepath.name)
        version_s.append(filename)
    return list(sorted(version_s))


def _revision_of(filename: str):
    return filename.rsplit("_", 1)[-1]


def _generate_ddl_upgrade(
    directory: MigrationDirectory, filename, prev_revision, revision
):
    target = directory.ddl / f"{filename}.upgrade.sql"
    if not target.exists():
        with _alembic_config_output(directory) as (cfg, output):
            revision_range = f"{prev_revision}:{revision}"
            _alembic.upgrade(cfg, sql=True, revision=revision_range)
            target.write_text(output.getvalue())


def _generate_ddl_downgrade(
    directory: MigrationDirectory, filename, prev_revision, revision
):
    target = directory.ddl / f"{filename}.rollback.sql"
    if not target.exists():
        with _alembic_config_output(directory) as (cfg, output):
            revision_range = f"{revision}:{prev_revision}"
            _alembic.downgrade(cfg, sql=True, revision=revision_range)
            target.write_text(output.getvalue())


def generate_ddl():
    directory = MigrationDirectory.from_base_dir()
    directory.ddl.mkdir(exist_ok=True)
    prev_revision = "base"
    for filename in _load_versions(directory):
        revision = _revision_of(filename)
        _generate_ddl_upgrade(directory, filename, prev_revision, revision)
        _generate_ddl_downgrade(directory, filename, prev_revision, revision)
        prev_revision = revision


def _current_revision(directory: MigrationDirectory) -> str:
    with _alembic_config_output(directory) as (cfg, output):
        _alembic.current(cfg)
        revision = output.getvalue().strip()
    if not revision:
        revision = "base"
    # alembic output example: 4ed04aaf7743 (head)
    revision = revision.split()[0]
    return revision


def migrate(target: str):
    directory = MigrationDirectory.from_base_dir()
    versions = _load_versions(directory)
    if not versions:
        raise CommandError("no migration versions")
    revision_map = {"base": (0, "base")}
    for idx, filename in enumerate(versions, 1):
        revision_map[_revision_of(filename)] = (idx, filename)
    if target == "head":
        target = list(revision_map.keys())[-1]
    target = _revision_of(target)
    if target not in revision_map:
        raise CommandError(f"target revision {target!r} not found")
    target_idx, target_version = revision_map[target]
    current = _current_revision(directory)
    assert current in revision_map, f"current revision {current!r} not found"
    current_idx, current_version = revision_map[current]
    LOG.info(f"Migrate {current_version} -> {target_version}")
    if target_idx == current_idx:
        return
    with _alembic_config_output(directory) as (cfg, _):
        if current_idx < target_idx:
            _alembic.upgrade(cfg, target)
        else:
            _alembic.downgrade(cfg, target)



import click

from .migration import _command


@click.group("sql")
def main():
    """EZTea SQL commands"""


@main.command()
@click.argument("app", type=str)
def init(app):
    """
    Init alembic migration configs.

    Example: eztea sql init myapp
    """
    _command.init(app)


@main.command()
def auto_revision():
    """Auto generate migration version."""
    _command.auto_revision()


@main.command()
def generate_ddl():
    """Generate DDL SQL of migration versions."""
    _command.generate_ddl()


@main.command()
@click.argument("target", type=str, default="head")
def migrate(target):
    """
    Update database schema to target revision.

    TARGET: revision identifier, base or head, default is head.
    """
    _command.migrate(target)



import sqlalchemy as sa
from sqlalchemy.dialects.mysql import LONGTEXT
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy_jsonfield import JSONField as _JSONField
from sqlalchemy_utc import UtcDateTime as _UtcDateTime

from eztea import json

__all__ = (
    "JSONField",
    "UtcDateTime",
)


class JSONField(_JSONField):
    __doc__ = _JSONField.__doc__

    # https://sqlalche.me/e/14/cprf
    cache_ok = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            enforce_string=True,
            enforce_unicode=True,
            json=json,
            *args,
            **kwargs
        )

    def load_dialect_impl(self, dialect: DefaultDialect) -> TypeEngine:
        """Select impl by dialect.

        :return: dialect implementation depends of decoding method
        :rtype: TypeEngine
        """
        if dialect.name == "mysql":
            return dialect.type_descriptor(LONGTEXT)
        return dialect.type_descriptor(sa.UnicodeText)


class UtcDateTime(_UtcDateTime):
    __doc__ = _UtcDateTime.__doc__
    # https://sqlalche.me/e/14/cprf
    cache_ok = True



import logging
import time
import typing
from contextlib import contextmanager
from threading import RLock
from urllib.parse import quote_plus

import sqlalchemy as sa
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy_utils import create_database, database_exists, drop_database

from ._timing import SqlalchemyConnectionTiming, TimingValue

LOG = logging.getLogger(__name__)

TableNamesOrMetaData = typing.Union[typing.List[str], sa.MetaData]


class SqlalchemyConnection:
    def __init__(
        self,
        *,
        db_type: str,
        db_name: str,
        db_user: str,
        db_password: str,
        db_host: str,
        db_port: int,
        pool_size: int = 5,
        pool_overflow: int = 0,
        pool_recycle: int = 600,
        pool_timeout: int = 5,
        echo_query: bool = False,
        echo_pool: bool = False,
    ):
        self._db_type = db_type
        self._db_user = db_user
        self._db_password = db_password
        self._db_host = db_host
        self._db_port = db_port
        self._db_name = db_name

        self._pool_size = pool_size
        self._pool_overflow = pool_overflow
        self._pool_recycle = pool_recycle
        self._pool_timeout = pool_timeout

        self._echo_query = echo_query
        self._echo_pool = echo_pool

        self._engine = None
        self._lock = RLock()
        self._timing = SqlalchemyConnectionTiming()

        self.url = self._get_url()

    def _get_url(self):
        url_getter = {
            "mysql": self._get_mysql_url,
            "postgresql": self._get_postgresql_url,
        }.get(self._db_type)
        if not url_getter:
            raise ValueError("not support db_type {}".format(self._db_type))
        return url_getter()

    @property
    def timing(self) -> TimingValue:
        return self._timing.value

    def reset_timing(self):
        self._timing.reset()

    def _get_mysql_url(self):
        return "mysql+pymysql://{user}:{password}@{host}:{port}/{name}?charset=utf8mb4".format(
            user=self._db_user,
            password=quote_plus(self._db_password),
            host=self._db_host,
            port=self._db_port,
            name=self._db_name,
        )

    def _get_postgresql_url(self):
        return "postgresql+psycopg2://{user}:{password}@{host}:{port}/{name}".format(
            user=self._db_user,
            password=quote_plus(self._db_password),
            host=self._db_host,
            port=self._db_port,
            name=self._db_name,
        )

    def is_database_exists(self) -> bool:
        return database_exists(self.url)

    def get_table_names(self) -> typing.List[str]:
        """
        Return all existed table names database.
        """
        inspector: Inspector = sa.inspect(self.get_engine())
        return list(inspector.get_table_names())

    def _sql_quote(self, value: str) -> str:
        return self.get_engine().dialect.identifier_preparer.quote(value)

    def create_database(self):
        """create database if not exists"""
        if not self.is_database_exists():
            encoding_map = {"mysql": "utf8mb4", "postgresql": "utf8"}
            create_database(self.url, encoding=encoding_map[self._db_type])

    def drop_database(self):
        """drop database if exists"""
        if self.is_database_exists():
            drop_database(self.url)

    def _get_meta_data_table_names(self, meta_data: sa.MetaData):
        return [x.name for x in reversed(meta_data.sorted_tables)]

    def _get_param_table_names(self, tables: TableNamesOrMetaData = None):
        if tables is None:
            return self.get_table_names()
        if isinstance(tables, sa.MetaData):
            return self._get_meta_data_table_names(tables)
        return tables

    def drop_all_tables(self, tables: TableNamesOrMetaData = None):
        """drop all tables in the database if exists"""
        table_names = self._get_param_table_names(tables)
        with self.atomic() as db:
            for table in table_names:
                sql = "DROP TABLE IF EXISTS {}".format(self._sql_quote(table))
                db.execute(sql)

    def clear_database(self, tables: TableNamesOrMetaData = None):
        """delete all records of tables in the database"""
        table_names = self._get_param_table_names(tables)
        with self.atomic() as db:
            for table in table_names:
                # For small table, DELETE is faster than TRUNCATE
                sql = "DELETE FROM {}".format(self._sql_quote(table))
                db.execute(sql)

    def _create_engine(self) -> Engine:
        # https://docs.sqlalchemy.org/en/13/core/pooling.html
        engine: Engine = sa.create_engine(
            self.url,
            pool_pre_ping=False,
            pool_size=self._pool_size,
            max_overflow=self._pool_overflow,
            pool_timeout=self._pool_timeout,
            pool_recycle=self._pool_recycle,
            echo=self._echo_query,
            echo_pool=self._echo_pool,
        )
        sa.event.listen(
            engine, "before_cursor_execute", self._before_cursor_execute
        )
        return engine

    def _before_cursor_execute(self, *args, **kwargs):
        self._timing.increase_count()

    def get_engine(self) -> Engine:
        if self._engine is None:
            with self._lock:
                if self._engine is None:
                    self._engine = self._create_engine()
        return self._engine

    def close(self):
        with self._lock:
            if self._engine is not None:
                self._engine.dispose()
                self._engine = None

    @contextmanager
    def connect(self) -> typing.Iterator[Connection]:
        with self.get_engine().connect() as conn:
            yield conn

    @contextmanager
    def atomic(self) -> typing.Iterator[Connection]:
        start_time = time.time()
        try:
            with self.connect() as conn:
                with conn.begin():
                    yield conn
        finally:
            self._timing.increase_cost(time.time() - start_time)

    def check(self) -> bool:
        try:
            with self.atomic() as db:
                db.execute("select 1").scalar()
        except Exception as ex:
            LOG.info(f"SQL connection check failed: {ex}", exc_info=ex)
            return False
        return True



from typing import Optional

from falcon import testing

import eztea.json as ezjson

from ._helper import join_url_path, shorten


class CallResult:
    def __init__(self, response: testing.Result):
        self._response = response

    def __repr__(self):
        if self.is_success:
            msg = self._response.status
        else:
            msg = f"{self.error}: {shorten(self.message, 80)}"
        return f"<{type(self).__name__} {msg}>"

    @property
    def raw(self) -> testing.Result:
        return self._response

    @property
    def is_success(self):
        return 200 <= self._response.status_code <= 299

    @property
    def data(self) -> dict:
        return self._response.json

    @property
    def error(self) -> Optional[str]:
        if self.is_success:
            return None
        try:
            return self.data["error"]
        except (ezjson.JSONDecodeError, KeyError, TypeError):
            return str(self._response.status_code)

    @property
    def message(self) -> Optional[str]:
        if self.is_success:
            return None
        try:
            return self.data["message"]
        except (ezjson.JSONDecodeError, KeyError, TypeError):
            return self._response.text


class WebTestClient:
    def __init__(self, app, headers=None, prefix: str = ""):
        self.__client = testing.TestClient(app, headers=headers)
        self.__prefix = prefix

    @property
    def raw(self) -> testing.TestClient:
        return self.__client

    def get(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a GET request to a WSGI application.
        """
        return self.request("GET", path, **kwargs)

    def head(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a HEAD request to a WSGI application.
        """
        return self.request("HEAD", path, **kwargs)

    def post(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a POST request to a WSGI application.
        """
        return self.request("POST", path, **kwargs)

    def put(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a PUT request to a WSGI application.
        """
        return self.request("PUT", path, **kwargs)

    def options(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate an OPTIONS request to a WSGI application.
        """
        return self.request("OPTIONS", path, **kwargs)

    def patch(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a PATCH request to a WSGI application.
        """
        return self.request("PATCH", path, **kwargs)

    def delete(self, path="/", **kwargs) -> testing.Result:
        """
        Simulate a DELETE request to a WSGI application.
        """
        return self.request("DELETE", path, **kwargs)

    def request(
        self, method="GET", path="/", *args, **kwargs
    ) -> testing.Result:
        """
        Simulate a request to a WSGI application.
        """
        path = join_url_path(self.__prefix, path)
        return self.__client.simulate_request(method, path, *args, **kwargs)

    def call(self, api: str, **kwargs) -> CallResult:
        response = self.post(f"/{api}", json=kwargs)
        return CallResult(response)
